# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import json
import os
import zipfile
from collections import OrderedDict

import pkgschema

from ..config import config
from ..utils import sign_dict


def dump_package(package, fn):
    """Dumps a package into a file."""
    with open(fn, 'w') as fp:
        json.dump(package, fp, indent=4, sort_keys=True)
        fp.write('\n')


def load_package(fn):
    """Loads a package from package dump file."""
    from .package import Package
    with open(fn) as fp:
        dump = json.load(fp, object_pairs_hook=OrderedDict)
    return Package(dump=dump)


def _generate_archive_name(package, output):
    if output is not None:
        return output
    return '{0.product}-{0.version}.uhupkg'.format(package)


def dump_package_archive(package, output=None, force=False):
    """Saves package as an archive. Returns genereted archive filename.

    Generated archive is a gz compressed tar file with current package
    metadata and all objects files.

    All objects are renamed to its hash and moved to the archive
    root. Objects are included without duplication and links are
    resolved.
    """
    # Checks minimum package requirements
    if package.version is None:
        raise ValueError('Cannot generate archive without package version.')
    if package.product is None:
        raise ValueError('Cannot generate archive without product UID.')
    if not package.objects.all():
        raise ValueError('Cannot generate archive without objects.')
    # Checks metadata complience
    metadata = package.to_metadata()
    try:
        pkgschema.validate_metadata(metadata)
    except pkgschema.ValidationError:
        raise ValueError('Cannot generate archive with invalid metadata.')

    # Checks archive output
    output = _generate_archive_name(package, output)
    if os.path.exists(output) and not force:
        raise FileExistsError('Archive "{}" already exists.'.format(output))

    # Writes archive
    cache = set()
    signature = sign_dict(metadata, config.get_private_key_path())
    metadata = json.dumps(metadata, sort_keys=True)
    with zipfile.ZipFile(output, mode='w') as archive:
        archive.writestr('signature', signature)
        archive.writestr('metadata', metadata)
        for obj in package.objects.all():
            sha256sum = obj['sha256sum']
            if sha256sum in cache:
                continue
            cache.add(sha256sum)
            archive.write(os.path.realpath(obj.filename), sha256sum)
    return output



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import click

from ..core._options import Options


TYPES = {
    'absolute_path': click.Path(readable=False),
    'boolean': click.BOOL,
    'integer': click.INT,
    'string': click.STRING,
}


class ClickObjectOption(click.Option):

    def __init__(self, option):
        if option.choices:
            type_ = click.Choice(option.choices)
        elif option.min is not None or option.max is not None:
            type_ = click.IntRange(min=option.min, max=option.max)
        else:
            type_ = TYPES[option.type_name]
        super().__init__(
            param_decls=option.cli, help=option.help,
            default=None, type=type_)
        self.metadata = option.metadata


def get_object_options():
    options = {}
    for option in Options.all():
        if not option.volatile and option != Options.get('filename'):
            click_option = ClickObjectOption(option)
            options[click_option.name] = click_option
    return options


CLICK_ADD_OPTIONS = get_object_options()



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import os
import re

from prompt_toolkit.validation import Validator

from .exceptions import ValidationError


# pylint: disable=too-few-public-methods
class ContainerValidator(Validator):

    def __init__(self, name, container, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.container = [str(elm) for elm in container]

    def validate(self, document):
        element = document.text.strip()
        if not element:
            message = '{} is required'.format(self.name)
            raise ValidationError(document, message)
        if element not in self.container:
            message = '"{}" is not a valid {}'.format(element, self.name)
            raise ValidationError(document, message)


# pylint: disable=too-few-public-methods
class ObjectUIDValidator(Validator):

    def validate(self, document):
        uid = document.text.split('#')[0].strip()
        if not uid:
            raise ValidationError(document, 'You must specify an object')
        if not uid.isdigit():
            message = '"{}" is not a valid object UID'.format(uid)
            raise ValidationError(document, message)


class ObjectOptionValueValidator(Validator):

    def __init__(self, option, mode, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mode = mode
        self.option = option
        self.document = None
        self.value = None

    def validate(self, document):
        self.document = document
        # Normalize
        self.value = document.text.strip()

        # Checks if a blank value is valid
        if not self.value:
            if self.option.default is not None:
                return None
            if self.mode.is_required(self.option):
                raise ValidationError(document, 'You must provide a value')

        # Validates against metadata rules
        try:
            self.validate_value()
        except ValueError as err:
            raise ValidationError(document, str(err))

    def validate_value(self):
        if self.option.metadata == 'filename':
            self.validate_filename()
        elif self.option.metadata == 'target-type':
            self.validate_target_type()
        else:
            self.option.validate(self.value)

    def validate_filename(self):
        if not os.path.exists(self.value):
            message = '"{}" does not exist'.format(self.value)
            raise ValidationError(self.document, message)
        if os.path.isdir(self.value):
            raise ValidationError(self.document, 'Only files are allowed')

    def validate_target_type(self):
        if self.value not in self.mode.target_types:
            message = 'You must pass a valid target type (eg. {})'.format(
                ', '.join(self.mode.target_types))
            raise ValidationError(self.document, message)


class PackageUIDValidator(Validator):

    def validate(self, document):
        value = document.text.strip()
        if not value:
            message = 'You need to specify a package UID'
            raise ValidationError(document, message)
        if not re.match(r'[a-fA-F0-9]{64}', value):
            message = 'You need specify a valid package UID'
            raise ValidationError(document, message)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import click

from .. import get_version
from ..repl import repl

from .config import config_cli, cleanup_command
from .hardware import hardware_cli
from .package import package_cli
from .product import product_cli


@click.group(invoke_without_command=True)
@click.option('--package', type=click.Path())
@click.version_option(
    get_version(), message='UpdateHub Utils - %(version)s')
@click.pass_context
def cli(ctx, package):
    """UpdateHub utility.

    To push packages, set USE_SERVER_URL environment variable to
    UpdateHub API server address.
    """
    if ctx.invoked_subcommand is None:
        repl(package)


# General commands
cli.add_command(cleanup_command)

# Subcommands
cli.add_command(config_cli)
cli.add_command(hardware_cli)
cli.add_command(package_cli)
cli.add_command(product_cli)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

# Keyword used to identify that all hardware identifiers are supported
ANY = 'any'

SUPPORTED_HARDWARE_ERROR = 'supported-hardware key has an invalid value.'


class SupportedHardwareManager:
    """Supported hardware manager."""

    # Metadata schema property name
    metadata = 'supported-hardware'

    def __init__(self, dump=None):
        if dump is None:
            self._hardware = set()
        else:
            self._hardware = self._from_dump(dump)

    def _from_dump(self, dump):
        value = dump.get(self.metadata)
        if value is None:
            raise ValueError(SUPPORTED_HARDWARE_ERROR)
        if value == ANY:
            return set()
        if isinstance(value, str):
            raise ValueError(SUPPORTED_HARDWARE_ERROR)
        try:
            return set(value)
        except TypeError:
            raise ValueError(SUPPORTED_HARDWARE_ERROR)

    def all(self):
        """Returns all supported hardware indentifiers alphabetically."""
        return sorted(self._hardware)

    def add(self, hardware):
        """Adds a new supported hardware identifier."""
        self._hardware.add(hardware)

    def remove(self, hardware):
        """Tries to remove a supported hardware identifier from the list.

        Raises KeyError if identifier does not exist."""
        try:
            self._hardware.remove(hardware)
        except KeyError:
            err = 'Hardware "{}" does not exist or is already removed.'
            raise KeyError(err.format(hardware))

    def reset(self):
        """Removes all hardware indentifiers from set."""
        self._hardware.clear()

    def to_metadata(self):
        """Serializes supported hardware as template."""
        if not self:
            return {self.metadata: ANY}
        return {self.metadata: self.all()}

    def to_template(self):
        """Serializes supported hardware as template."""
        return self.to_metadata()

    def __eq__(self, other):
        return self.to_metadata() == other.to_metadata()

    def __iter__(self):
        return iter(self.all())

    def __len__(self):
        return len(self._hardware)

    def __str__(self):
        hardware = ', '.join(self.all()) or ANY
        return 'Supported hardware: {}'.format(hardware)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import requests

from ..utils import get_custom_ca_certs_file
from ._request import Request, HTTPError


UNKNOWN_ERROR = 'A unexpected request error ocurred. Try again later.'


def request(method, url, *args, sign=True, **kwargs):
    custom_ca_certs_file = get_custom_ca_certs_file()

    if custom_ca_certs_file is not None and 'verify' not in kwargs:
        kwargs['verify'] = custom_ca_certs_file

    try:
        if sign:
            response = Request(url, method, *args, **kwargs).send()
        else:
            response = requests.request(
                method, url, *args, timeout=30, **kwargs)
    except HTTPError as error:
        raise error
    except (requests.exceptions.MissingSchema,
            requests.exceptions.InvalidSchema,
            requests.exceptions.URLRequired,
            requests.exceptions.InvalidURL):
        raise HTTPError('You have provided an invalid server URL.')
    except requests.ConnectTimeout:
        raise HTTPError('Connection timed out. Try again later.')
    except requests.ConnectionError:
        raise HTTPError('Server is not available. Try again later.')
    except requests.RequestException:
        raise HTTPError(UNKNOWN_ERROR)
    if response.status_code == 401:
        raise HTTPError('Unautorized. Did you set your credentials?')
    elif not response.ok:
        raise HTTPError(format_server_error(response))
    return response


def get(url, *args, **kwargs):
    return request('GET', url, *args, **kwargs)


def post(url, *args, **kwargs):
    return request('POST', url, *args, **kwargs)


def put(url, *args, **kwargs):
    return request('PUT', url, *args, **kwargs)


def format_server_error(response):
    try:
        body = response.json()
    except ValueError:
        return UNKNOWN_ERROR

    message = body.get('error_message')
    if message:
        return message

    errors = body.get('errors')
    if errors:
        try:
            return '\n'.join(
                ['- {}: {}'.format(k, ', '.join(v))
                 for k, v in errors.items()])
        except (AttributeError, TypeError):
            pass

    return UNKNOWN_ERROR



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

from copy import deepcopy

from ._options import OptionType, Options
from .install_condition import normalize_install_if_different


def normalize(obj, values):
    """Normalizes (clean and convert) all options.

    This is the first step of validation. We remove all options with
    None values. Then, we check if the remaning options are valid in
    this mode. Finally, we convert the value to what is expected on
    metadata.
    """
    cleaned = {}
    for option, value in values.items():
        option = Options.get(option)
        if value is not None:  # removes null values
            if option not in obj.options:  # checks allowed options
                err = '{} is a invalid option for {} mode'
                raise ValueError(err.format(option, obj.mode))
            cleaned[option] = option.validate(value)  # converts value
    return cleaned


def inject_default_values(obj, values):
    """Adds default values for all missing options."""
    for option in obj.options:
        values = inject_default_value(obj, option, values)
    return values


def inject_default_value(obj, option, values):
    """Adds the default value for a given option.

    The trickiest part in this process is that we cannot add
    values if its requirements are not satisfied. Since
    requirements may also have requirements and its requirements
    may also have default values, this problem is recursive.

    Some cases covered here:

    1. option is already presented: must do nothing

    2. option has no default: must do nothing

    3. option has default but no requirements: must insert
    default.

    4. option has default and requirements: must insert default
    and recursively insert requirements defaults.

    For all cases with injection, we must validate if the
    injection is possible (if the actual option requirements are
    satisfied). In case of invalid data, we must return the
    original values.
    """
    # We need a copy of the passed values since dicts are passed
    # by reference between function calls and we need be able to
    # return the original values in case of invalid data.
    original_values = deepcopy(values)

    # Option default value injection step
    if option not in values and option.default is not None:
        values[option] = option.default
    else:
        return original_values

    # Requirements default values injection step (recursive part)
    for req in option.requirements:
        values = inject_default_value(obj, req, values)

    # Validation step
    try:
        validate_option_requirements(option, values)
    except ValueError:
        return original_values
    return values


def validate_required_options(obj, values):
    """Checks if all mode required options are present."""
    for option in obj.required_options:
        if option not in values:
            err = 'Option "{}" is required for mode "{}".'
            raise ValueError(err.format(option, obj.mode))


def validate_options_requirements(values):
    """Verifies if all options requirements are satisfied."""
    for option in values:
        validate_option_requirements(option, values)


def validate_option_requirements(option, values):
    """Verifies if option requirements are satisfied."""
    # validate values argument
    for key in values.keys():
        if not isinstance(key, OptionType):
            err = 'values argument must have OptionType keys type (got {}).'
            raise TypeError(err.format(type(key)))

    if not option.requirements:
        return

    for req_option, req_value in option.requirements.items():
        if req_option not in values:
            err = ('You must specify a value for "{}" '
                   'when using "{}" options.')
            raise ValueError(err.format(req_option, option))
        value = values.get(req_option)
        if value != req_value:
            err = ('"{}" must be equal to "{}" when using "{}" '
                   'option. Got "{}".')
            raise ValueError(
                err.format(req_option, req_value, option, value))


def validate_options(obj, values):
    """Performs full object validation"""
    values = normalize_install_if_different(values)
    values = normalize(obj, values)
    values = inject_default_values(obj, values)
    validate_required_options(obj, values)
    validate_options_requirements(values)
    return values



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import click

from .utils import error, open_package


@click.group(name='hardware')
def hardware_cli():
    """Supported hardware related commands."""


@hardware_cli.command(name='add')
@click.argument('hardware')
def add_supported_hardware(hardware):
    """Adds a hardware identifier to supported hardware list."""
    with open_package() as package:
        package.supported_hardware.add(hardware)


@hardware_cli.command(name='remove')
@click.argument('hardware')
def remove_supported_hardware(hardware):
    """Removes a hardware identifier from supported hardware list."""
    with open_package() as package:
        try:
            package.supported_hardware.remove(hardware)
        except KeyError as err:
            error(2, err)


@hardware_cli.command(name='reset')
def reset_supported_hardware_list():
    """Removes all supported hardware identifiers."""
    with open_package() as package:
        package.supported_hardware.reset()



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

import json

import click

from pkgschema import validate_metadata, ValidationError

from ..core.object import Modes
from ..updatehub.api import get_package_status, UpdateHubError
from ..core.utils import dump_package, dump_package_archive
from ..ui import get_callback, show_cursor

from ._object import CLICK_ADD_OPTIONS
from .utils import error, open_package


@click.group(name='package')
def package_cli():
    """Package related commands."""


@package_cli.command('version')
@click.argument('version')
def set_version_command(version):
    """Sets package version."""
    with open_package() as package:
        package.version = version


@package_cli.command('show')
def show_command():
    """Shows all configured objects."""
    with open_package(read_only=True) as package:
        print(package)


@package_cli.command('export')
@click.argument('filename', type=click.Path(dir_okay=False))
def export_command(filename):
    """Copy package file to the given filename."""
    with open_package() as package:
        dump_package(package.to_template(with_version=False), filename)


# Object commands

@package_cli.command('add')
@click.argument('filename', type=click.Path(exists=True))
@click.option('--mode', '-m', type=click.Choice(Modes.names()),
              help='How the object will be installed', required=True)
def add_object_command(filename, mode, **options):
    """Adds an entry in the package file for the given artifact."""
    options = {CLICK_ADD_OPTIONS[opt].metadata: value
               for opt, value in options.items()
               if value is not None}
    options['filename'] = filename
    options['mode'] = mode
    with open_package() as package:
        try:
            package.objects.create(options)
        except ValueError as err:
            error(2, err)


# Adds all object options into cmd
for opt in CLICK_ADD_OPTIONS.values():
    add_object_command.params.append(opt)


@package_cli.command(name='edit')
@click.option('--index', type=click.INT, required=True,
              help='The object index')
@click.option('--installation-set', type=click.INT,
              help='The installation set to retrive object')
@click.option('--option', help='The object option to be edited', required=True)
@click.option('--value', help='The new value to be set', required=True)
def edit_object_command(index, installation_set, option, value):
    """Edits an object property within package."""
    with open_package() as package:
        try:
            package.objects.update(
                index, option, value, set_index=installation_set)
        except ValueError as err:
            error(3, err)


@package_cli.command('remove')
@click.argument('object-id', type=click.INT)
def remove_object_command(object_id):
    """Removes the filename entry within package file."""
    with open_package() as package:
        package.objects.remove(object_id)


# Transaction commands

@package_cli.command(name='push')
def push_command():
    """Pushes a package file to server with the given version."""
    callback = get_callback()
    with open_package(read_only=True) as package:
        try:
            package.push(callback)
        except UpdateHubError as err:
            error(2, err)
        finally:
            show_cursor()


@package_cli.command(name='status')
@click.argument('package-uid')
def status_command(package_uid):
    """Prints the status of the given package."""
    try:
        print(get_package_status(package_uid))
    except UpdateHubError as err:
        error(2, err)


@package_cli.command(name='metadata')
def metadata_command():
    """Loads package and prints its metadata."""
    with open_package(read_only=True) as package:
        metadata = package.to_metadata()
        print(json.dumps(metadata, indent=4, sort_keys=True))
    try:
        validate_metadata(metadata)
        print('Valid metadata.')
    except ValidationError as err:
        error(1, err)


@package_cli.command(name='archive')
@click.option('--output', type=click.Path(dir_okay=False),
              help="Where to write archive")
@click.option('--force', is_flag=True,
              help="Overwrites output file if output exists")
def archive_command(output, force):
    """Saves package as archive."""
    with open_package(read_only=True) as package:
        try:
            dump_package_archive(package, output, force)
        except FileExistsError as err:
            error(1, err)
        except ValueError as err:
            error(2, err)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

# options must be loaded before modes. Also it must be at package
# load level so we can register all classes before anything else.
from . import options
from . import modes
from .package import Package



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

from setuptools import find_packages, setup

from uhu import get_version


setup(
    name='uhu',
    description="UpdateHub's firmware update package management utilities",
    keywords='industrial-linux embedded-linux embedded firmware-updates linux update-service',  # nopep8
    version=get_version(),
    packages=find_packages(exclude=['tests*']),
    entry_points={
        'console_scripts': ['uhu=uhu.cli:cli']
    },
    install_requires=[
        'click>=6.5',
        'humanize>=0.5.1',
        'progress>=1.1',
        'prompt-toolkit>=2.0.0,<3.0.0',
        'pycrypto',
        'updatehub-package-schema>=1.0.3',
        'requests>=2.20.0',
        'rfc3987>=1.3',
    ],
    author='O.S. Systems Software LTDA',
    author_email='contato@ossystems.com.br',
    url='http://www.ossystems.com.br',
    license='GPL-2.0',
    zip_safe=False,
)



# Copyright (C) 2017 O.S. Systems Software LTDA.
# SPDX-License-Identifier: GPL-2.0

from ._object import BaseObject


class CopyObject(BaseObject):
    mode = 'copy'
    allow_compression = True
    allow_install_condition = True
    target_types = ['device']
    options = [
        'filename',
        'size',
        'sha256sum',
        'target-type',
        'target',
        'target-path',
        'filesystem',
        'mount-options',
        'format?',
        'format-options',
    ]
    required_options = [
        'filename',
        'target-type',
        'target',
        'target-path',
        'filesystem',
    ]
    requirements = [
        ('format?', 'format-options', True)
    ]
    string_template = [
        ('target', ('filesystem',)),
        ('format?', ('format-options',)),
        ('mount-options', ()),
        ('target-path', ())
    ]


class FlashObject(BaseObject):
    mode = 'flash'
    allow_compression = False
    allow_install_condition = True
    target_types = ['device', 'mtdname']
    options = [
        'filename',
        'sha256sum',
        'size',
        'target-type',
        'target',
    ]
    required_options = [
        'filename',
        'target-type',
        'target',
    ]
    string_template = [
        ('target', ()),
    ]


class IMXKobsObject(BaseObject):
    mode = 'imxkobs'
    allow_compression = False
    allow_install_condition = True
    options = [
        'filename',
        'size',
        'sha256sum',
        '1k_padding',
        'search_exponent',
        'chip_0_device_path',
        'chip_1_device_path',
    ]
    required_options = [
        'filename',
    ]
    string_template = [
        ('1k_padding', ()),
        ('search_exponent', ()),
        ('chip_0_device_path', ()),
        ('chip_1_device_path', ()),
    ]


class RawObject(BaseObject):
    mode = 'raw'
    allow_compression = True
    allow_install_condition = True
    target_types = ['device']
    options = [
        'filename',
        'size',
        'sha256sum',
        'target-type',
        'target',
        'chunk-size',
        'count',
        'seek',
        'skip',
        'truncate',
    ]
    required_options = [
        'filename',
        'target-type',
        'target',
    ]
    string_template = [
        ('target', ('seek', 'truncate')),
        ('chunk-size', ()),
        ('skip', ()),
        ('count', ()),
    ]


class TarballObject(BaseObject):
    mode = 'tarball'
    allow_compression = False
    allow_install_condition = False
    target_types = ['device', 'mtdname', 'ubivolume']
    options = [
        'filename',
        'size',
        'sha256sum',
        'target-type',
        'target',
        'target-path',
        'filesystem',
        'mount-options',
        'format?',
        'format-options',
    ]
    required_options = [
        'filename',
        'target-type',
        'target',
        'target-path',
        'filesystem',
    ]
    string_template = [
        ('target', ('filesystem',)),
        ('format?', ('format-options',)),
        ('mount-options', ()),
        ('target-path', ())
    ]


class UBIFSObject(BaseObject):
    mode = 'ubifs'
    allow_compression = True
    allow_install_condition = False
    target_types = ['ubivolume']
    options = [
        'filename',
        'size',
        'sha256sum',
        'target-type',
        'target',
    ]
    required_options = [
        'filename',
        'target-type',
        'target',
    ]
    string_template = [
        ('target', ()),
    ]


class MenderObject(BaseObject):
    mode = 'mender'
    allow_compression = False
    allow_install_condition = False
    options = [
        'filename',
        'sha256sum',
        'size',
    ]
    required_options = [
        'filename',
    ]


class ZephyrObject(BaseObject):
    mode = 'zephyr'
    allow_compression = False
    allow_install_condition = False
    options = [
        'filename',
        'sha256sum',
        'size',
    ]
    required_options = [
        'filename',
    ]
